@@ -50,22 +50,22 @@ struct _GnomeDesktopFileParser {
 #define VALID_KEY_CHAR 1
 #define VALID_LOCALE_CHAR 2
 guchar valid[256] = { 
-   0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 
-   0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 
-   0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x3 , 0x2 , 0x0 , 
-   0x3 , 0x3 , 0x3 , 0x3 , 0x3 , 0x3 , 0x3 , 0x3 , 0x3 , 0x3 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 
-   0x0 , 0x3 , 0x3 , 0x3 , 0x3 , 0x3 , 0x3 , 0x3 , 0x3 , 0x3 , 0x3 , 0x3 , 0x3 , 0x3 , 0x3 , 0x3 , 
-   0x3 , 0x3 , 0x3 , 0x3 , 0x3 , 0x3 , 0x3 , 0x3 , 0x3 , 0x3 , 0x3 , 0x0 , 0x0 , 0x0 , 0x0 , 0x2 , 
-   0x0 , 0x3 , 0x3 , 0x3 , 0x3 , 0x3 , 0x3 , 0x3 , 0x3 , 0x3 , 0x3 , 0x3 , 0x3 , 0x3 , 0x3 , 0x3 , 
-   0x3 , 0x3 , 0x3 , 0x3 , 0x3 , 0x3 , 0x3 , 0x3 , 0x3 , 0x3 , 0x3 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 
-   0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 
-   0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 
-   0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 
-   0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 
-   0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 
-   0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 
-   0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 
-   0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 
+  /* 0   */ 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 
+  /* 16  */ 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 
+  /* 32  */ 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x3 , 0x2 , 0x0 , 
+  /* 48  */ 0x3 , 0x3 , 0x3 , 0x3 , 0x3 , 0x3 , 0x3 , 0x3 , 0x3 , 0x3 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 
+  /* 64  */ 0x2 , 0x3 , 0x3 , 0x3 , 0x3 , 0x3 , 0x3 , 0x3 , 0x3 , 0x3 , 0x3 , 0x3 , 0x3 , 0x3 , 0x3 , 0x3 , 
+  /* 80  */ 0x3 , 0x3 , 0x3 , 0x3 , 0x3 , 0x3 , 0x3 , 0x3 , 0x3 , 0x3 , 0x3 , 0x0 , 0x0 , 0x0 , 0x0 , 0x2 , 
+  /* 96  */ 0x0 , 0x3 , 0x3 , 0x3 , 0x3 , 0x3 , 0x3 , 0x3 , 0x3 , 0x3 , 0x3 , 0x3 , 0x3 , 0x3 , 0x3 , 0x3 , 
+  /* 112 */ 0x3 , 0x3 , 0x3 , 0x3 , 0x3 , 0x3 , 0x3 , 0x3 , 0x3 , 0x3 , 0x3 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 
+  /* 128 */ 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 
+  /* 144 */ 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 
+  /* 160 */ 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 
+  /* 176 */ 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 
+  /* 192 */ 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 
+  /* 208 */ 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 
+  /* 224 */ 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 
+  /* 240 */ 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0 , 0x0
 };
 
 static void                     report_error   (GnomeDesktopFileParser   *parser,
@@ -508,7 +508,8 @@ parse_key_value (GnomeDesktopFileParser *parser, GError **error)
   gchar *value_start;
   gchar *value;
   gchar *p;
-
+  char *key;
+  
   line_end = strchr (parser->line, '\n');
   if (line_end == NULL)
     line_end = parser->line + strlen (parser->line);
@@ -595,7 +596,12 @@ parse_key_value (GnomeDesktopFileParser *parser, GError **error)
     }
 
   line = new_line (parser);
-  line->key = g_quark_from_static_string (g_strndup (key_start, key_end - key_start));
+  key = g_strndup (key_start, key_end - key_start);
+  line->key = g_quark_try_string (key);
+  if (line->key == 0)
+    line->key = g_quark_from_static_string (key);
+  else
+    g_free (key);
   if (locale_start)
     line->locale = g_strndup (locale_start, locale_end - locale_start);
   line->value = value;
@@ -1384,7 +1390,19 @@ gnome_desktop_file_merge_string_into_list (GnomeDesktopFile *df,
   {
     /* Append to current list */
     char *str;
-    str = g_strconcat (raw ? raw : "", value, ";", NULL);
+    if (raw)
+      {
+        int len;
+        len = strlen (raw);
+        /* handle broken file with no ';' after the list */
+        if (len > 0 && raw[len-1] != ';')
+          str = g_strconcat (raw, ";", value, ";", NULL);
+        else
+          str = g_strconcat (raw, value, ";", NULL);
+      }
+    else
+      str = g_strconcat (value, ";", NULL);
+    
     gnome_desktop_file_set_raw (df, section, keyname, locale, str);
     g_free (str);
   }
@@ -1403,7 +1421,7 @@ gnome_desktop_file_remove_string_from_list (GnomeDesktopFile *df,
   if (gnome_desktop_file_get_strings (df, section, keyname, locale,
                                       &values, &n_values))
     {
-      int i;
+      int i, j;
       int n_found;
 
       n_found = 0;
@@ -1411,25 +1429,41 @@ gnome_desktop_file_remove_string_from_list (GnomeDesktopFile *df,
       i = 0;
       while (i < n_values)
         {
-          if (n_found > 0)
+          if (values[i] == NULL)
+            continue;
+
+          if (strcmp (values[i], value) == 0)
             {
               g_free (values[i]);
-              values[i] = values[i+1];
-              values[i+1] = NULL;
-            }
-
-          if (((i+1) < n_values) &&
-              strcmp (values[i], value) == 0)
-            {
+              values[i] = NULL;
               ++n_found;
             }
-          else
-            {
-              ++i;
-            }
+
+          ++i;
         }
 
-      if (n_found > 0)
+      i = j = 0;
+      while (i < n_values)
+        {
+          j = MAX (i + 1, j);
+          if (values[i] == NULL)
+            {
+              while (j < n_values && values[j] == NULL)
+                j++;
+
+              if (j < n_values)
+                {
+                  values[i] = values[j];
+                  values[j] = NULL;
+                }
+            }
+
+          ++i;
+        }
+
+      if (n_found == n_values)
+        gnome_desktop_file_unset (df, section, keyname);
+      else if (n_found > 0)
         gnome_desktop_file_set_strings (df, section, keyname, locale,
                                         (const char**) values);
       
@@ -1445,15 +1479,15 @@ struct {
   const char *encoding;
   const char *langs[N_LANG];
 } known_encodings[] = {
-  {"ARMSCII-8", {"by"}},
+  {"ARMSCII-8", {"hy"}},
   {"BIG5", {"zh_TW"}},
   {"CP1251", {"be", "bg"}},
-  {"EUC-CN", {"zh_TW"}},
+  {"EUC-CN", {"zh_CN"}},
   {"EUC-JP", {"ja"}},
   {"EUC-KR", {"ko"}},
   {"GEORGIAN-ACADEMY", {}},
   {"GEORGIAN-PS", {"ka"}},
-  {"ISO-8859-1", {"br", "ca", "da", "de", "en", "es", "eu", "fi", "fr", "gl", "it", "nl", "wa", "no", "pt", "pt", "sv"}},
+  {"ISO-8859-1", {"br", "ca", "da", "de", "en", "es", "eu", "fi", "fr", "gl", "it", "nl", "no", "pt", "sv", "wa" }},
   {"ISO-8859-2", {"cs", "hr", "hu", "pl", "ro", "sk", "sl", "sq", "sr"}},
   {"ISO-8859-3", {"eo"}},
   {"ISO-8859-5", {"mk", "sp"}},
@@ -1518,13 +1552,13 @@ static const char *
 get_canonical_encoding (const char *encoding)
 {
   int i;
-  for (i = 0; i < G_N_ELEMENTS (enc_aliases); i++)
+  for (i = 0; i < (int) G_N_ELEMENTS (enc_aliases); i++)
     {
       if (aliases_equal (enc_aliases[i].alias, encoding))
 	return enc_aliases[i].value;
     }
 
-  for (i = 0; i < G_N_ELEMENTS (known_encodings); i++)
+  for (i = 0; i < (int) G_N_ELEMENTS (known_encodings); i++)
     {
       if (aliases_equal (known_encodings[i].encoding, encoding))
 	return known_encodings[i].encoding;
@@ -1555,7 +1589,7 @@ get_encoding_from_lang (const char *lang)
 {
   int i, j;
   
-  for (i = 0; i < G_N_ELEMENTS (known_encodings); i++)
+  for (i = 0; i < (int) G_N_ELEMENTS (known_encodings); i++)
     {
       for (j = 0; j < N_LANG; j++)
 	{
